module M100Util where

import Char
import Parse
import qualified Bag as B

import Data.List
import Control.Monad
import Data.Either
import Data.Maybe
import qualified Data.Set as S
import System.IO.SSC.Haskell98 as SSC

charsFromString :: [Namespace] -> String -> ([String], [Character])
charsFromString pref = partitionEithers . mapMaybe parse . filter (not . comment_line) . lines
  where
    comment_line [] = True
    comment_line xs = lc == '-' || lc == '@'
      where
        lc = last xs
    parse ln = case parseCharacterG pref $ dropWhile (==' ') ln of
      Just x -> Just (Right x)
      Nothing
        | looksLikeCharacterSpec ln -> Just (Left ln)
        | otherwise -> Nothing

readCharsFromFileWithIgnored :: [Namespace] -> FilePath -> IO ([String], [Character])
readCharsFromFileWithIgnored pref fp = fmap (charsFromString pref) $ SSC.readFile fp

readCharsFromFile :: [Namespace] -> FilePath -> IO [Character]
readCharsFromFile pref fp = liftM snd $ readCharsFromFileWithIgnored pref fp

uniquifyNames :: [Character] -> [Character]
uniquifyNames chs = zipWith replace chs $ uniquify $ map chName chs
  where
    uniquify xs = snd $ mapAccumL f (B.toSet b) xs
      where
        b = B.fromList xs
        f used x
          | B.count x b > 1 = (S.insert x' used, x')
          | otherwise = (used, x)
          where
            x' = head $ filter (not . flip S.member used) $ map ((x++) . ('_':) . show) [0..]
    replace ch name = ch{ chName = name }
