# ifndef M100SIM_INT_H
# define M100SIM_INT_H

/*# define USE_GMP_INTEGER*/

struct danger_of_overflow{};

# ifdef USE_GMP_INTEGER

#  include "gmpxx.h"

typedef mpz_class int_t;
typedef const mpz_class &intparam;
inline void print_int(FILE *out, intparam z) { mpz_out_str(out, 10, z.get_mpz_t()); }
inline int_t div_floor(intparam x, intparam y)
{
  mpz_class ret;
  mpz_fdiv_q(ret.get_mpz_t(), x.get_mpz_t(), y.get_mpz_t());
  return ret;
}
inline int_t checked_double(intparam x) { return 2 * x; }
inline int_t checked_subtract(intparam x, intparam y) { return x - y; }
inline bool small_enough(int x) { return true; }

# else

#  include <stdio.h>

typedef int int_t;
typedef int intparam;
inline void print_int(FILE *out, intparam z) { fprintf(out, "%d", z); }
inline int_t div_floor(intparam x, intparam y)
{
  if(x >= 0) return x / y;
  return - ((-x-1+y) / y);
}
inline int_t checked_double(intparam x)
{
  if(x < (1 << 29) && 1 > -(1 << 29)) return x * 2;
  throw danger_of_overflow();
}
inline int_t checked_subtract(intparam x, intparam y)
{
  if(x > -(1 << 29) && y < (1 << 29)) return x - y;
  throw danger_of_overflow();
}
inline bool small_enough(int x) { return x < (1 << 30); }

# endif

# endif
