import JSON
import RoundRobin
import Char
import Match
import GameUtil
import M100Util

import qualified Data.Map as M
import System.IO.SSC.Haskell98 as SSC
import System.Environment

data MatchResult = MR
  { mrCharNumber :: (Int, Int)
  , mrWinner :: Maybe TeamRef
  , mrLog :: String
  }

main = do
  SSC.putStrLn ""
  [def] <- getArgs
  let 
    ss = case def of
      "void" -> mvoidNamespaces
      "1a" -> m2011aNamespaces
      _ -> m100Namespaces
  (ignored, chs0) <- fmap (charsFromString ss) $ SSC.getContents
  let chs = uniquifyNames chs0
  (mrs, tbl) <- getResults chs
  SSC.print $ toDoc $ Object $ M.fromList
    [ ("chars", Array $ map (String . show) chs)
    , ("names", Array $ map (String . chName) chs)
    , ("matches", Array $ map mrToJSON mrs)
    , ("table", Array $ map (Array . map (String . showWinner)) tbl)
    , ("nwin", Array $ map (Number . count (Just Team0)) tbl)
    , ("nlose", Array $ map (Number . count (Just Team1)) tbl)
    , ("ndraw", Array $ map (Number . count Nothing) tbl)
    , ("ignored", Array $ map String ignored)
    ]
  where
    showWinner w = case w of
      Nothing -> "×"
      Just Team0 -> "○"
      Just Team1 -> "●"

    count x = fromIntegral . length . filter (==x)

mrToJSON :: MatchResult -> Value
mrToJSON MR{ mrCharNumber = (x, y), mrLog = log } = Object $ M.fromList
  [ ("x", Number $ fromIntegral x)
  , ("y", Number $ fromIntegral y)
  , ("log", String log)
  ]

getResults :: [Character] -> IO ([MatchResult], [[Maybe TeamRef]])
getResults cs = do
  mr_table <- roundRobin (zip cs [0..]) $ \(y, yn) (x, xn) -> let
    (gr, msgs, gs) = match x y
    in seq (grWinner gr) $ return MR{ mrCharNumber = (xn, yn), mrWinner = fmap opponent $ grWinner gr, mrLog = unlines msgs }
  let simple_table = completeTable Nothing (fmap opponent) $ map (map mrWinner) mr_table
  return (concat mr_table, simple_table)
