-- #hide
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental
   
   This module contains some definitions convinient for defining
   memory IO ports.
-}
module System.IO.SSC.Internal.BasicPort
  ( basicPort
  , BasicPort
  , bpD
  , bpBase
  , bpLock
  , bpTryLock
  , bpUnsafeUnlock
  , bpIsClosed
  , bpClose
  , bpEnsureOpen
  ) where

import Control.Monad
import Control.Concurrent.MVar
import Data.Maybe
import Data.IORef

data BasicPort b = BP {-# UNPACK #-} !(MVar ()) {-# UNPACK #-} !(IORef Bool) !b

basicPort :: b -> IO (BasicPort b)
basicPort b = do
  l <- newMVar ()
  c <- newIORef False
  return $ BP l c b

bpBase :: BasicPort b -> b
bpBase (BP _ _ b) = b

bpD :: (b -> r) -> BasicPort b -> r
bpD k = k . bpBase

bpGetLocker :: BasicPort b -> IO (MVar ())
bpGetLocker (BP l _ _) = return l

bpLock :: BasicPort b -> IO ()
bpLock (BP l _ _) = takeMVar l

bpTryLock :: BasicPort b -> IO Bool
bpTryLock (BP l _ _) = liftM isJust $ tryTakeMVar l

bpUnsafeUnlock :: BasicPort b -> IO ()
bpUnsafeUnlock (BP l _ _) = putMVar l ()

bpIsClosed :: BasicPort b -> IO Bool
bpIsClosed (BP _ c _) = readIORef c

bpClose :: BasicPort b -> (b -> IO ()) -> IO ()
bpClose (BP _ c b) close = do
  close b
  writeIORef c True

bpEnsureOpen :: BasicPort b -> IO ()
bpEnsureOpen (BP _ c _) = do
  closed <- readIORef c
  when closed $ ioError $ userError "port is closed"
