{-# OPTIONS_GHC -fglasgow-exts -fallow-undecidable-instances #-}
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

   A Char Stream -> Char Stream transformer that automatically
   flushes the underlying buffers when a newline is being output.
   This is not an efficient implementation, so avoid using it
   when possible.
-}
module System.IO.SSC.LineBuffering
  ( oLineBuffering
  , ioLineBuffering
  , LineBuffering
  ) where

import System.IO.SSC.Classes
import System.IO.SSC.BufferSize

newtype LineBuffering d pos ba ma s = LB s

oLineBuffering :: s -> IO (LineBuffering d pos ba ma s)
oLineBuffering = lineBuffering

ioLineBuffering :: s -> IO (LineBuffering d pos ba ma s)
ioLineBuffering = lineBuffering

lineBuffering :: s -> IO (LineBuffering d pos ba ma s)
lineBuffering s = return (LB s)

instance (Port s) => Port (LineBuffering d pos ba ma s) where
  rawClose (LB s) = rawClose s
  rawIsClosed (LB s) = rawIsClosed s
  rawFlush (LB s) = rawFlush s

instance (Destination d s) => Destination d (LineBuffering d pos ba ma s) where
  rawDest (LB s) = rawDest s

instance (Lockable s) => Lockable (LineBuffering d pos ba ma s) where
  sLock (LB s) = sLock s
  sTryLock (LB s) = sTryLock s
  sUnsafeUnlock (LB s) = sUnsafeUnlock s

instance (Discardable s) => Discardable (LineBuffering d pos ba ma s) where
  rawDiscard (LB s) = rawDiscard s

instance (Seekable pos s) => Seekable pos (LineBuffering d pos ba ma s) where
  rawGetPos (LB s) = rawGetPos s
  rawSetPos (LB s) = rawSetPos s

instance (RelativeSeekable pos s) => RelativeSeekable pos (LineBuffering d pos ba ma s) where
  rawRelativeSeek (LB s) = rawRelativeSeek s

instance (SizedInput pos s) => SizedInput pos (LineBuffering d pos ba ma s) where
  rawInputEnd (LB s) = rawInputEnd s
  rawSeekFromEnd (LB s) = rawSeekFromEnd s

instance (SizedOutput pos s) => SizedOutput pos (LineBuffering d pos ba ma s) where
  rawOutputEnd (LB s) = rawOutputEnd s

instance (IStream Char s) => IStream Char (LineBuffering d pos ba ma s) where
  rawGet (LB s) = rawGet s
  rawGets (LB s) = rawGets s
  rawGetContents (LB s) = rawGetContents s

instance (RetrievableStream Char s) => RetrievableStream Char (LineBuffering d pos ba ma s) where
  rawGetUntil (LB s) = rawGetUntil s

instance (OStream Char s) => OStream Char (LineBuffering d pos ba ma s) where
  rawPut (LB s) '\n' = rawPut s '\n' >> rawFlush s
  rawPut (LB s) c = rawPut s c
  rawPuts (LB s) [] = return ()
  rawPuts (LB s) cs = case splitAt defaultBufferSize cs of
    (init, rest) -> if '\n' `elem` init
      then do
        rawPuts s init
        rawPuts s rest
        rawFlush s
      else do
        rawPuts s init
        rawPuts (LB s) rest

instance (IBlock ba s) => IBlock ba (LineBuffering d pos ba ma s) where
  rawGetBufB (LB s) = rawGetBufB s

instance (OBlock ba s) => OBlock ba (LineBuffering d pos ba ma s) where
  rawPutBufB (LB s) = rawPutBufB s

instance (IMemory ma s) => IMemory ma (LineBuffering d pos ba ma s) where
  rawReadMem (LB s) = rawReadMem s
  rawEndRead (LB s) = rawEndRead s
  rawFillMemB (LB s) = rawFillMemB s

instance (OMemory ma s) => OMemory ma (LineBuffering d pos ba ma s) where
  rawWriteMem (LB s) = rawWriteMem s
  rawEndWrite (LB s) = rawEndWrite s
  rawFlushMemB (LB s) = rawFlushMemB s
