{-# OPTIONS_GHC -fffi -cpp -optP-P #-}
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

   A Word8 memory -> Char buffer transformer which performs
   locale-dependent encoding conversion.
-}
module System.IO.SSC.LocaleEncoding
  ( iLocaleEncoding
  , oLocaleEncoding
  , ioLocaleEncoding
  , iNaturalEncoding
  , oNaturalEncoding
  , ioNaturalEncoding
  , initCtypeLocale
  , ILocaleEncoding
  , OLocaleEncoding
  , IOLocaleEncoding
  ) where

import System.IO.SSC.Classes
import System.IO.SSC.Internal.Fd
import System.IO.SSC.Internal.Debug
# ifdef mingw32_HOST_OS
import System.IO.SSC.UTF16Block
import System.IO.SSC.Win32Unicode
import System.IO.SSC.Convert
import System.IO.SSC.Buffer
import Foreign.C.Types
# else
import System.IO.SSC.Iconv
# endif
import Data.Word
import Foreign.C.String

foreign import ccall unsafe "locale_encoding.h init_ctype" init_ctype :: IO ()

-- | Initialize LC_CTYPE locale with the value that is determined by
-- the system environment. This function just calls 
-- @setlocale(LC_CTYPE, \"\")@.
initCtypeLocale :: IO ()
initCtypeLocale = init_ctype

# ifdef mingw32_HOST_OS

foreign import ccall unsafe "locale_encoding.h locale_cp" locale_cp :: IO CUInt
foreign import ccall unsafe "locale_encoding.h locale_input_console_cp" input_console_cp :: IO CUInt
foreign import ccall unsafe "locale_encoding.h locale_output_console_cp" output_console_cp :: IO CUInt

type LocaleEncoding d pos a m cb bb = cb d pos (bb d pos Word16 (cb d pos m a Word8 Word16)) a Word16 Char
type ILocaleEncoding d pos a m = LocaleEncoding d pos a m IConversionBlock IBuffer
type OLocaleEncoding d pos a m = LocaleEncoding d pos a m OConversionBlock OBuffer
type IOLocaleEncoding d pos a m = LocaleEncoding d pos a m IOConversionBlock IOBuffer

loc f mem = do
  cp <- locale_cp
  f (fromIntegral cp) mem

nat f mem fd getcp = do
  tty <- fdIsTerminal fd
  cp <- if tty
    then getcp
    else locale_cp
  f (fromIntegral cp) mem

iLocaleEncoding :: m -> IO (ILocaleEncoding d pos a m)
iLocaleEncoding base = loc iMb2wBlock base >>= iBuffer >>= iUtf16Block

oLocaleEncoding :: m -> IO (OLocaleEncoding d pos a m)
oLocaleEncoding base = loc oMb2wBlock base >>= oBuffer >>= oUtf16Block

ioLocaleEncoding :: m -> IO (IOLocaleEncoding d pos a m)
ioLocaleEncoding base = loc ioMb2wBlock base >>= ioBuffer >>= ioUtf16Block

iNaturalEncoding :: Fd -> m -> IO (ILocaleEncoding d pos a m)
iNaturalEncoding fd base = 
  nat iMb2wBlock base fd input_console_cp >>= iBuffer >>= iUtf16Block

oNaturalEncoding :: Fd -> m -> IO (OLocaleEncoding d pos a m)
oNaturalEncoding fd base =
  nat oMb2wBlock base fd output_console_cp >>= oBuffer >>= oUtf16Block

ioNaturalEncoding :: Fd -> m -> IO (IOLocaleEncoding d pos a m)
ioNaturalEncoding _ base = ioLocaleEncoding base

# else

foreign import ccall unsafe "locale_encoding.h locale_encoding" locale_encoding :: IO CString

type ILocaleEncoding d pos a m = IIconvBlock d pos m a Word8 Char
type OLocaleEncoding d pos a m = OIconvBlock d pos m a Word8 Char
type IOLocaleEncoding d pos a m = IOIconvBlock d pos m a Word8 Char

-- | Transform Word8 memory port into Char block port. This function
-- respects the current LC_CTYPE locale. You probably want to
-- set the locale before calling this function.
localeEncoding :: (String -> m -> IO t) -> m -> IO t
localeEncoding iconvChar mem = do
  enc <- locale_encoding >>= peekCString
  iconvChar enc mem

iLocaleEncoding :: m -> IO (ILocaleEncoding d pos a m)
iLocaleEncoding = localeEncoding iIconvChar

oLocaleEncoding :: m -> IO (OLocaleEncoding d pos a m)
oLocaleEncoding = localeEncoding oIconvChar

ioLocaleEncoding :: m -> IO (IOLocaleEncoding d pos a m)
ioLocaleEncoding = localeEncoding ioIconvChar

iNaturalEncoding :: Fd -> m -> IO (ILocaleEncoding d pos a m)
iNaturalEncoding _ = iLocaleEncoding

oNaturalEncoding :: Fd -> m -> IO (OLocaleEncoding d pos a m)
oNaturalEncoding _ = oLocaleEncoding

ioNaturalEncoding :: Fd -> m -> IO (IOLocaleEncoding d pos a m)
ioNaturalEncoding _ = ioLocaleEncoding

# endif
