{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

   Conversion between UTF-16 string and [Char].
-}
module System.IO.SSC.UTF16Block
  ( utf16ToCharConverter
  , charToUTF16Converter
  , iUtf16Block
  , oUtf16Block
  , ioUtf16Block
  ) where

import System.IO.SSC.Classes
import System.IO.SSC.Convert
import Control.Monad
import Data.Bits
import Data.Char
import Data.Word
import Foreign.Ptr
import Foreign.Storable
import System.IO.Error

utf16ToCharConverter :: Converter Word16 Char
utf16ToCharConverter iptr isize optr osize = loop 0 0
  where
    loop read written
      | read == isize || written == osize = return (read, written)
      | otherwise = do
        wc <- peekElemOff iptr read
        if wc < 0xd800 || 0xe000 <= wc
          then do
            pokeElemOff optr written (chr $ fromIntegral wc)
            loop (read + 1) (written + 1)
          else do
            when (0xdc00 <= wc) ilseq
            if read + 1 == isize
              then return (read, written)
              else do
                wc2 <- peekElemOff iptr (read + 1)
                when (wc2 < 0xdc00 || 0xe000 <= wc2) ilseq
                pokeElemOff optr written (chr $ fromIntegral $ shiftL (wc - 0xd800) 10 .|. (wc2 - 0xdc00) .|. 0x100000)
                loop (read + 2) (written + 1)
    ilseq = ioeIlseq "System.IO.SSC.UTF16Block.utf16ToCharConverter"

charToUTF16Converter :: Converter Char Word16
charToUTF16Converter iptr isize optr osize = loop 0 0
  where
    loop read written
      | read == isize || written == osize = return (read, written)
      | otherwise = do
        c <- peekElemOff iptr read
        when ('\xd800' <= c && c < '\xe000') ilseq
        if c < '\x10000'
          then do
            pokeElemOff optr written (fromIntegral $ ord c)
            loop (read + 1) (written + 1)
          else if written + 1 == osize
            then return (read, written)
            else do
              let v = ord c
              pokeElemOff optr written (fromIntegral $ (shiftR v 10 .&. 0x3ff) .|. 0xd800)
              pokeElemOff optr (written + 1) (fromIntegral $ (v .&. 0x3ff) .|. 0xdc00)
              loop (read + 1) (written + 2)
    ilseq = ioeIlseq "System.IO.SSC.UTF8Block.charToUTF16Converter"

ioeIlseq :: String -> IO a
ioeIlseq desc = ioError $ mkIOError illegalOperationErrorType desc Nothing Nothing

iUtf16Block :: m -> IO (IConversionBlock d pos m a Word16 Char)
iUtf16Block mem = iConversionBlock utf16ToCharConverter mem

oUtf16Block :: m -> IO (OConversionBlock d pos m a Word16 Char)
oUtf16Block mem = oConversionBlock charToUTF16Converter mem

ioUtf16Block :: m -> IO (IOConversionBlock d pos m a Word16 Char)
ioUtf16Block mem = ioConversionBlock utf16ToCharConverter charToUTF16Converter mem
